/**
 * @fileoverview Externs for binary-tools 2.0
 * @see {@link https://rochars.com/binary-tools}
 * @externs
 * @license
 * Copyright (c) 2023-2025 Rafael da Silva Rocha.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

// #FIXME hack to GCC understand Error.cause
/** @type {Object} */
RangeError.prototype.cause = RangeError.prototype.cause || {values: []};
/** @type {Object} */
TypeError.prototype.cause = RangeError.prototype.cause || {values: []};

/** @type {!Object} */
var btools = {};

// packing/unpacking

/**
 * Pack values to a byte buffer.
 * @param {string} format The struct format definition.
 * @param {!Array<*>} values The values to pack.
 * @param {!Array<number>} buffer The byte buffer to write on.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to start writing.
 * @return {number} The next index to write on the buffer.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} If the output buffer is typed array and size is not valid.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
btools.packTo = function(format, values, buffer, clamp=false, index=0) {};

/**
 * Unpack values from an array of bytes to an Array-like object.
 * Always start writing the output at the beginning of the output array.
 * @param {string} format The struct format definition.
 * @param {!Array<number>} buffer The byte buffer to unpack.
 * @param {!Array<number|string|bigint|boolean>} output The output Array.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to read.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
btools.unpackTo = function(format, buffer, output, clamp=false, index=0) {};

/**
 * Pack values as a array of bytes.
 * @param {string} format The struct format definition.
 * @param {!Array<*>} values The values to pack.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type on the format string.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
btools.pack = function(format, values, clamp=false) {};

/**
 * Unpack values from an array of bytes.
 * This method returns an Array even if only a single value is unpacked.
 * @param {string} format The struct format definition.
 * @param {!Array<number>} buffer The byte buffer to unpack.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to read.
 * @return {!Array<number|string|bigint|boolean>} The unpacked values.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
btools.unpack = function(format, buffer, clamp=false, index=0) {};

// writing/reading

/**
 * Write a array of values to a byte buffer.
 * @param {string} format The format definition.
 * @param {!Array<*>} values The values to write.
 * @param {!Array<number>} buffer The buffer to write on.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to start writing.
 * @return {number} The next buffer index to write on.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} If the output buffer is typed array and size is not valid.
 * @throws {RangeError} On integer overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
btools.writeTo = function(format, values, buffer, clamp=false, index=0) {};

/**
 * Read a array of values from a byte buffer to a array or a typed array.
 * @param {string} format The format definition.
 * @param {!Array<number>} buffer The byte buffer to read.
 * @param {!Array<number|string|bigint|boolean>} output The output array.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [start=0] The input buffer index to start reading.
 * @param {number=} [end=buffer.length] The input buffer index to stop reading.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
btools.readTo = function(
    format, buffer, output, clamp=false, start=0, end=buffer.length) {};

/**
 * Write a array of values as a array of bytes.
 * @param {string} format The format definition.
 * @param {!Array<*>} values The values to write.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type on the format string.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
btools.write = function(format, values, clamp=false) {};

/**
 * Read a array of values from a byte buffer.
 * @param {string} format The format definition.
 * @param {!Array<number>} buffer The byte buffer.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [start=0] The buffer index to start reading.
 * @param {number=} [end=buffer.length] The buffer index to stop reading.
 * @return {!Array<number|string|bigint|boolean>}
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
btools.read = function (
  format, buffer, clamp=false, start=0, end=buffer.length) {};

// tools

/**
 * Swap the byte ordering in a buffer. The buffer is modified in place.
 * @param {!Array<number>} buffer The bytes.
 * @param {number} offset The byte offset.
 * @param {number=} [start=0] The start index.
 * @param {number=} [end=bytes.length] The end index.
 * @function endianness
 * @memberOf module:binary-tools
 */
btools.endianness = function(buffer, offset, start=0, end=buffer.length) {};

/**
 * Set a type in the types object.
 * @param {string} typeSymbol The type symbol.
 * @param {number} bits The type bits.
 * @param {boolean=} [signed=false] True for signed, false otherwise.
 * @throws {Error} On unsupported type.
 */
btools.setType = function(typeSymbol, bits, signed=false) {};

/**
 * Calculate the buffer size based on a format string.
 * @param {string} format The format string.
 * @return {number}
 * @throws {Error} On unsupported type.
 * @function calcSize
 * @memberOf module:binary-tools
 */
btools.calcSize = function(format) {};

/**
 * Returns how many bytes are needed to serialize a UTF-8 string.
 * @param {string} str The string to pack.
 * @return {number} The number of bytes needed for the string.
 */
btools.utf8BufferSize = function(str) {};

/**
 * Returns how many bytes are needed to serialize a UTF-16 string.
 * @param {string} str The string to pack.
 * @param {boolean=} [forceBOM=false] If BOM should be enforced or not.
 * If false (default), then it will only count the bytes according to
 * the characters on the string; if true and string have no BOM, then
 * it will count the bytes of the characters plus BOM (size + 2).
 * @return {number} The number of bytes needed for the string.
 */
btools.utf16BufferSize = function(str, forceBOM=false) {};

/**
 * Returns how many bytes are needed to serialize a UTF-32 string.
 * @param {string} str The string to pack.
 * @param {boolean=} [forceBOM=false] If BOM should be enforced or not.
 * If false (default), then it will only count the bytes according to
 * the characters on the string; if true and string have no BOM, then
 * it will count the bytes of the characters plus BOM (size + 4).
 * @return {number} The number of bytes needed for the string.
 */
btools.utf32BufferSize = function(str, forceBOM=false) {};

/**
 * Format a byte array as a string of hex numbers.
 * @param {!Array<number>} bytes The bytes.
 * @return {string}
 * @throws {RangeError} If bytes contains invalid values.
 */
btools.bytesToHex = function(bytes) {};

/**
 * Convert a hex string to a byte array.
 * @param {string} hexStr The hex string.
 * @return {!Array<number>}
 * @throws {RangeError} If string contains chars outside the 0..f range.
 * @throws {Error} If string length is not even.
 */
btools.hexToBytes = function(hexStr) {};

// types

/**
 * The standard binary-tools datatype definitions.
 * New integer types can be created with setType().
 * For the type name, numbers and the characters \@~=<>! are reserved, and also
 * all characters used for pre-defined types (xcCsSuUVNbB?hHegtTlLfjJkKdqQAYOM).
 * @type {Object.<string, Object>}
 * @property {DataType} x Pad byte
 * @property {DataType} c ASCII character
 * @property {DataType} C Extended ASCII (ISO-8859-1) character
 * @property {DataType} s ASCII string
 * @property {DataType} S Extended ASCII (ISO-8859-1) string
 * @property {DataType} u UTF-8 string
 * @property {DataType} U UTF-16 string
 * @property {DataType} V UTF-32 string
 * @property {DataType} N Unsiged 4-bit integers
 * @property {DataType} b Signed 8-bit integers
 * @property {DataType} B Unsigned 8-bit integers
 * @property {DataType} ? Boolean
 * @property {DataType} h Signed 16-bit integers
 * @property {DataType} H Unsigned 16-bit integers
 * @property {DataType} e 16-bit Half-precision floating-point numbers
 * @property {DataType} g 16-bit Brain floating point numbers
 * @property {DataType} t Signed 24-bit integers
 * @property {DataType} T Unsigned 24-bit integers
 * @property {DataType} i Signed 32-bit integers
 * @property {DataType} I Unsigned 32-bit integers
 * @property {DataType} l Signed 32-bit integers
 * @property {DataType} L Unsigned 32-bit integers
 * @property {DataType} f 32-bit Single-precision floating-point numbers
 * @property {DataType} j Signed 40-bit integers
 * @property {DataType} J Unsigned 40-bit integers
 * @property {DataType} k Signed 48-bit integers
 * @property {DataType} K Unsigned 48-bit integers
 * @property {DataType} d 64-bit Double-precision floating-point numbers
 * @property {DataType} q signed 64-bit integers
 * @property {DataType} Q Unsigned 64-bit integers
 * @property {DataType} A Unsigned 128-bit integers
 * @property {DataType} Y Unsigned 256-bit integers
 * @property {DataType} O Unsigned 512-bit integers
 * @property {DataType} M Unsigned 1024-bit integers
 * @memberOf module:binary-tools
 */
btools.types = {
  'x': {},
  'c': {},
  'C': {},
  's': {},
  'S': {},
  'u': {},
  'U': {},
  'V': {},
  'N': {},
  'b': {},
  'B': {},
  '?': {},
  'h': {},
  'H': {},
  'e': {},
  'g': {},
  't': {},
  'T': {},
  'i': {},
  'I': {},
  'l': {},
  'L': {},
  'f': {},
  'j': {},
  'J': {},
  'k': {},
  'K': {},
  'd': {},
  'q': {},
  'Q': {},
  'A': {},
  'Y': {},
  'O': {},
  'M': {}
};
