// Type definitions for binary-tools 2.0
// Project: https://rochars.com/binary-tools
// Definitions by: Rafael da Silva Rocha <rocha.rafaelsilva@gmail.com>

// packing/unpacking

/**
 * Pack values to a byte buffer.
 * @param {string} format The struct format definition.
 * @param {!Array<*>} values The values to pack.
 * @param {!Array<number>} buffer The byte buffer to write on.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to start writing.
 * @return {number} The next index to write on the buffer.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} If the output buffer is typed array and size is not valid.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
export function packTo(
  format: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  buffer: ArrayLike<number>,
  clamp?: boolean,
  index?: number): number;

/**
 * Unpack values from an array of bytes to an Array-like object.
 * Always start writing the output at the beginning of the output array.
 * @param {string} format The struct format definition.
 * @param {!Array<number>} buffer The byte buffer to unpack.
 * @param {!Array<number|string|bigint|boolean>} output The output Array.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to read.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
export function unpackTo(
  format: string,
  buffer: ArrayLike<number>,
  output: ArrayLike<number|string|bigint>,
  clamp?: boolean,
  index?: number): void;

/**
 * Pack values as a array of bytes.
 * @param {string} format The struct format definition.
 * @param {!Array<*>} values The values to pack.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type on the format string.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
export function pack(
  format: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  clamp?: boolean): Array<number>;

/**
 * Unpack values from an array of bytes.
 * This method returns an Array even if only a single value is unpacked.
 * @param {string} format The struct format definition.
 * @param {!Array<number>} buffer The byte buffer to unpack.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to read.
 * @return {!Array<number|string|bigint|boolean>} The unpacked values.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
export function unpack(
  format: string,
  buffer: ArrayLike<number>,
  clamp?: boolean,
  index?: number): Array<number|string|bigint>;

// reading/writing

/**
 * Write a array of values to a byte buffer.
 * @param {string} format The format definition.
 * @param {!Array<*>} values The values to write.
 * @param {!Array<number>} buffer The buffer to write on.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [index=0] The buffer index to start writing.
 * @return {number} The next buffer index to write on.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} If the output buffer is typed array and size is not valid.
 * @throws {RangeError} On integer overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
export function writeTo(
  format: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  buffer: ArrayLike<number>,
  clamp?: boolean,
  index?: number): number;

/**
 * Read a array of values from a byte buffer to a array or a typed array.
 * @param {string} format The format definition.
 * @param {!Array<number>} buffer The byte buffer to read.
 * @param {!Array<number|string|bigint|boolean>} output The output array.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [start=0] The input buffer index to start reading.
 * @param {number=} [end=buffer.length] The input buffer index to stop reading.
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
export function readTo(
  format: string,
  buffer: ArrayLike<number>,
  output: ArrayLike<number|string|bigint>,
  clamp?: boolean,
  start?: number,
  end?: number): void;

/**
 * Write a array of values as a array of bytes.
 * @param {string} format The format definition.
 * @param {!Array<*>} values The values to write.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type on the format string.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If a value is not valid for its type.
 */
export function write(
  format: string,
  values: number|string|bigint|ArrayLike<number|string|bigint>,
  clamp?:boolean): Array<number>;

/**
 * Read a array of values from a byte buffer.
 * @param {string} format The format definition.
 * @param {!Array<number>} buffer The byte buffer.
 * @param {boolean=} [clamp=false] True to clamp values on overflow.
 * @param {number=} [start=0] The buffer index to start reading.
 * @param {number=} [end=buffer.length] The buffer index to stop reading.
 * @return {!Array<number|string|bigint|boolean>}
 * @throws {Error} On unsupported type on the format string.
 * @throws {Error} On bad input buffer length.
 */
export function read(
  format: string,
  buffer: ArrayLike<number>,
  clamp?: boolean,
  start?: number,
  end?: number): Array<number|string|bigint>;

// tools

/**
 * Swap the byte ordering in a buffer. The buffer is modified in place.
 * @param {!Array<number>} buffer The bytes.
 * @param {number} offset The byte offset.
 * @param {number=} [start=0] The start index.
 * @param {number=} [end=bytes.length] The end index.
 * @function endianness
 * @memberOf module:binary-tools
 */
export function endianness(
  buffer: ArrayLike<number>,
  offset: number,
  start?: number,
  end?: number): void;

/**
 * Set a type in the types object.
 * @param {string} typeSymbol The type symbol.
 * @param {number} bits The type bits.
 * @param {boolean=} [signed=false] True for signed, false otherwise.
 * @throws {Error} On unsupported type.
 */
export function setType(
  typeSymbol: string,
  bits: number,
  signed?: boolean): void;

/**
 * Calculate the buffer size based on a format string.
 * @param {string} format The format string.
 * @return {number}
 * @throws {Error} On unsupported type on the format string.
 */
export function calcSize(
  format: string): number;

/**
 * Returns how many bytes are needed to serialize a UTF-8 string.
 * @param {string} str The string to pack.
 * @return {number} The number of bytes needed for the string.
 */
export function utf8BufferSize(
  str: string): number

/**
 * Returns how many bytes are needed to serialize a UTF-16 string.
 * @param {string} str The string to pack.
 * @param {boolean=} [forceBOM=false] If BOM should be enforced or not.
 * If false (default), then it will only count the bytes according to
 * the characters on the string; if true and string have no BOM, then
 * it will count the bytes of the characters plus BOM (size + 2).
 * @return {number} The number of bytes needed for the string.
 */
export function utf16BufferSize(
  str: string,
  forceBOM?: boolean): number

/**
 * Returns how many bytes are needed to serialize a UTF-32 string.
 * @param {string} str The string to pack.
 * @param {boolean=} [forceBOM=false] If BOM should be enforced or not.
 * If false (default), then it will only count the bytes according to
 * the characters on the string; if true and string have no BOM, then
 * it will count the bytes of the characters plus BOM (size + 4).
 * @return {number} The number of bytes needed for the string.
 */
export function utf32BufferSize(
  str: string,
  forceBOM?: boolean): number

/**
 * Format a byte array as a string of hex numbers.
 * @param {!Array<number>} bytes The bytes.
 * @return {string}
 * @throws {RangeError} If bytes contains invalid values.
 */
export function bytesToHex(
  bytes: ArrayLike<number>): string

/**
 * Convert a hex string to a byte array.
 * @param {string} hexStr The hex string.
 * @return {!Array<number>}
 * @throws {RangeError} If string contains chars outside the 0..f range.
 * @throws {Error} If string length is not even.
 */
export function hexToBytes(
  hexStr: string): Array<number>

// types
